#
# CMX: Common Metadata eXchange
#
# (C)opyright 2024 cKnowledge.org
#

from cmeta.core import CMeta

from cmeta.core import access
from cmeta.core import error
from cmeta.core import halt

from cmeta.cli import cli

__version__ = '0.0.0.1'
